from django import template
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.templatetags.static import static
from django.urls import reverse
from django.utils.html import format_html
from django.utils.module_loading import import_string

from wagtail.admin.views.pages.preview import PreviewOnEdit
from wagtail.core.models import Page

try:
    # Wagtail >= 4.0
    from wagtail.core.models import Revision as PageRevision
except:
    from wagtail.core.models import PageRevision

import json

# monkey-patch PageRevision.as_page_object to store revision id on pages
original_as_page_object = PageRevision.as_page_object
def as_page_object(self, *args, **kwargs):
    page = original_as_page_object(self, *args, **kwargs)
    page._live_edit_revision_id = self.id
    return page
PageRevision.as_page_object = as_page_object

if hasattr(PreviewOnEdit, 'get_object'): # Wagtail >= 4.0
    # monkey-patch PreviewOnEdit.get_object so we can tell if we're looking at a unsaved preview
    original_get_object = PreviewOnEdit.get_object
    def get_object(self, *args, **kwargs):
        page = original_get_object(self, *args, **kwargs)
        page._live_edit_is_preview = True
        return page
    PreviewOnEdit.get_object = get_object
else:
    # monkey-patch PreviewOnEdit.get_page so we can tell if we're looking at a unsaved preview
    original_get_page = PreviewOnEdit.get_page
    def get_page(self, *args, **kwargs):
        page = original_get_page(self, *args, **kwargs)
        page._live_edit_is_preview = True
        return page
    PreviewOnEdit.get_page = get_page

register = template.Library()

is_enabled = import_string(getattr(settings, 'LIVEEDIT_ENABLED_CHECK', 'liveedit.utils.is_enabled'))

def is_authenticated(request):
    if request and request.user and request.user.is_authenticated:
        return True
    return False

@register.simple_tag(takes_context=True)
def liveedit_css(context):
    request = context.get('request')
    if not is_enabled(request) or not is_authenticated(request):
        return ''
    return format_html(
        '<link rel="stylesheet" type="text/css" href="{}">',
        static('css/liveedit.css')
    )

@register.simple_tag(takes_context=True)
def liveedit_js(context):
    request = context.get('request')
    if not is_enabled(request) or not is_authenticated(request):
        return ''
    return format_html(
        '<script type="text/javascript" src="{}"></script>',
        static('js/liveedit.js')
    )

@register.simple_tag(takes_context=True)
def liveedit_attributes(context):
    if 'liveedit_data' not in context:
        return ''

    if not context['liveedit_data'].get('id'):
        return ''

    return format_html('data-liveedit="{}"', 
        json.dumps(context['liveedit_data'])
    )

@register.simple_tag(takes_context=True)
def liveedit_include_block(context, block, object=None, field=None):
    context = context.flatten()
    request = context.get('request')

    def finish():
        return block.render_as_block(context)

    if not is_enabled(request) or not is_authenticated(request):
        return finish()

    if object and isinstance(object, Page):
        perms = object.permissions_for_user(request.user)
        if not perms.can_edit():
            return finish()

    data = {
        **(context.get('liveedit_data') or {}), 
        'id':block.id,
        'block_type': block.block_type
    }

    if object and field:
        if isinstance(object, Page):
            if object.has_unpublished_changes:
                # There is a new unpublished version of this page, are we looking at it?
                cur_rev_id = getattr(object, '_live_edit_revision_id', None)
                latest_rev_id = object.get_latest_revision().id

                if cur_rev_id!=latest_rev_id:
                    # We're not, so don't allow live editing, but send a link to the latest revision.
                    return finish() + format_html("<script>window._live_edit_draft_url='{}';</script>",
                            reverse('wagtailadmin_pages:revisions_view', args=(object.id, latest_rev_id))
                        )

                if getattr(object, '_live_edit_is_preview', False):
                    # We're viewing an unsaved preview, don't allow editing
                    return finish()

            elif getattr(request, 'is_dummy', False):
                # We are in preview mode, without a unpublished revision, don't allow editing
                return finish()

        data.update({
            'content_type_id': ContentType.objects.get_for_model(object).id,
            'object_id': object.id,
            'object_field': field
        })

    context['liveedit_data'] = data
    
    return finish()



from django import forms
from django.contrib.auth.decorators import permission_required
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.forms.utils import ErrorList
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.views.decorators.csrf import csrf_exempt

from wagtail.contrib.modeladmin.helpers import PermissionHelper

import wagtail
if wagtail.VERSION >= (3,):
    from wagtail.blocks import BlockWidget
    from wagtail.models import Page
    from wagtail.blocks.stream_block import StreamValue
else:
    from wagtail.core.blocks import BlockWidget
    from wagtail.core.models import Page
    from wagtail.core.blocks.stream_block import StreamValue

from collections.abc import Sequence
import json
import re
from urllib.parse import urlencode, urlparse

def islist(v):
    return isinstance(v, Sequence)

def isdict(v):
    return isinstance(v, dict)

def _find_block(stream_value, raw_data, block_id):
    for i in range(len(raw_data)):
        if raw_data[i]['id']==block_id:
            def set_value(val):
                stream_value[i].value = val
            return stream_value[i].block, stream_value[i].value, set_value, stream_value

        # is this a streamblock/listblock?
        if islist(raw_data[i]['value']) and len(raw_data[i]['value']) and hasattr(raw_data[i]['value'][0], 'get') and raw_data[i]['value'][0].get('id'):
            rb, rv, rsv, rp = _find_block(stream_value[i].value, raw_data[i]['value'], block_id)
            if rb:
                return rb, rv, rsv, rp

        #check structblock field values for any potential block lists
        if isdict(raw_data[i]['value']):
            for k, v in raw_data[i]['value'].items():
                if islist(v) and len(v) and hasattr(v[0], 'get') and v[0].get('id'):
                    rb, rv, rsv, rp = _find_block(stream_value[i].value[k], v, block_id)
                    if rb:
                        return rb, rv, rsv, rp

    return None, None, None, None

def find_block(stream_value, block_id):
    #returns Block, StreamValue.StreamChild, setter function to update value, parent StreamValue

    return _find_block(stream_value, stream_value.raw_data, block_id)

def modify_block(action, blocks, block_id):
    for i in range(len(blocks)):
        if action=="move_up" and i>0 and blocks[i]['id']==block_id:
            #swap with previous
            blocks[i-1], blocks[i] = blocks[i], blocks[i-1]
            return True #did the move
        elif action=="move_down" and i<(len(blocks)-1) and blocks[i]['id']==block_id:
            #swap with next
            blocks[i+1], blocks[i] = blocks[i], blocks[i+1]
            return True #did the move

        #is this a list?
        if islist(blocks[i]['value']):
            for v in blocks[i]['value']:
                if islist(v) and len(v) and hasattr(v[0], 'get') and v[0].get('id'):
                    if modify_block(action, v, block_id):
                        #if block was moved, stop looking
                        return True

        #search through all fields for any potential block lists
        if isdict(blocks[i]['value']):
            for v in blocks[i]['value'].values():
                if islist(v) and len(v) and hasattr(v[0], 'get') and v[0].get('id'):
                    if modify_block(action, v, block_id):
                        #if block was moved, stop looking
                        return True

def get_latest_revision_and_save_function(obj):
    """
    For a given model instance, return the latest revision of that model
    instance, as well as a function which saves it, as a tuple.
    """
    save = obj.save
    if isinstance(obj, Page):
        if obj.has_unpublished_changes:
            # Save over the latest revision, not the published one
            obj = obj.get_latest_revision_as_page()
            
            def save():
                rev = obj.get_latest_revision()
                rev.content_json = obj.to_json()
                rev.save()
    return obj, save

def check_can_edit(user, obj):
    if hasattr(obj, 'permissions_for_user'):
        if not obj.permissions_for_user(user).can_edit():
            raise PermissionDenied
    elif not PermissionHelper(obj).user_can_edit_obj(user, obj):
        raise PermissionDenied

def render_edit_panel(request, d):
    # Steal all of the <script> and stylesheet tags from the admin base template
    admin_base = render_to_string("wagtailadmin/admin_base.html", request=request)
    script_tags = mark_safe("\n".join(re.findall('<script[^>]*>.*?</script>', admin_base, re.DOTALL)))
    stylesheet_tags = mark_safe("\n".join(re.findall('<link rel="stylesheet"[^>]+/?>', admin_base, re.DOTALL)))

    try:
        editor_css = mark_safe(render_to_string("wagtailadmin/pages/_editor_css.html", request=request))
    except TemplateDoesNotExist as e:
        # Wagtail >= 4.0
        editor_css = ""

    ret = render(request, "liveedit/edit_panel.html", {
        **d,
        'script_tags': script_tags,
        'stylesheet_tags': stylesheet_tags,
        'editor_css': editor_css
    })
    ret['X-Frame-Options'] = 'SAMEORIGIN'
    return ret

def ReloadResponse(jump_to_id=None):
    msg = {'action':"reload", 'jump_to_id':jump_to_id}
    ret = HttpResponse('''
    <script>window.parent.postMessage(''' + json.dumps(msg) + ''', '*');</script>
    ''')
    ret['X-Frame-Options'] = 'SAMEORIGIN'
    return ret

# wrapper to workaround Wagtail's assumption that all blocks have a top level StreamBlock
class ErrorWrapper:
    def __init__(self, err):
        self.err = err
    def as_data(self):
        return self.err


@csrf_exempt
@permission_required('wagtailadmin.access_admin', login_url='wagtailadmin_login')
def action_view(request):
    ct = ContentType.objects.get(pk=request.POST['content_type_id'])
    obj = ct.model_class().objects.get(pk=request.POST['object_id'])
    check_can_edit(request.user, obj)
    obj, save = get_latest_revision_and_save_function(obj)
    value = getattr(obj, request.POST['object_field'])

    block_id = request.POST['id']

    modify_block(request.POST['action'], value.raw_data, block_id)
    save()

    return HttpResponseRedirect(request.POST['redirect_url'])

@permission_required('wagtailadmin.access_admin', login_url='wagtailadmin_login')
def edit_block_view(request):
    from wagtail.admin.views.pages.edit import EditView
    view = EditView.as_view()

    ct = ContentType.objects.get(pk=request.GET['content_type_id'])
    obj = ct.model_class().objects.get(pk=request.GET['object_id'])
    check_can_edit(request.user, obj)
    obj, save = get_latest_revision_and_save_function(obj)
    value = getattr(obj, request.GET['object_field'])

    block, block_value, set_value, parent = find_block(value, request.GET['id'])

    errors = ErrorWrapper(None)
    if request.method=="POST" and request.POST.get('delete'):
        for i, block in enumerate(parent):
            if block_value==block.value:
                del parent[i]
                save()

                return ReloadResponse()
    elif request.method=="POST":
        val = block.value_from_datadict(request.POST, request.FILES, 'block_edit_form')
        try:
            val = block.clean(val)

            set_value(val)
            #print('val is', repr(val))
            #print('block_value is', repr(block_value))

            save()

            return ReloadResponse(request.GET['id'])

        except forms.ValidationError as e:
            errors = ErrorWrapper(e)
            block_value = val

    bw = BlockWidget(block)

    return render_edit_panel(request, {
        'form_html': bw.render_with_errors('block_edit_form', block_value, errors=errors), #block.render_form(block_value, prefix="block_edit_form", ),
        'form_media': bw.media,
    })

@permission_required('wagtailadmin.access_admin', login_url='wagtailadmin_login')
def append_block_view(request):
    ct = ContentType.objects.get(pk=request.GET['content_type_id'])
    obj = ct.model_class().objects.get(pk=request.GET['object_id'])
    check_can_edit(request.user, obj)
    obj, save = get_latest_revision_and_save_function(obj)
    value = getattr(obj, request.GET['object_field'])

    _, _, _, parent_value = find_block(value, request.GET['id'])

    parent_block = parent_value.stream_block
    blank_value = StreamValue(parent_value.stream_block, [])

    errors = None
    if request.method=="POST":
        val = parent_block.value_from_datadict(request.POST, request.FILES, 'block_edit_form')
        try:
            val = parent_block.clean(val)

            for i, v in enumerate(parent_value):
                if v.id==request.GET['id']:
                    #insert after this block
                    for j, added in enumerate(val):
                        parent_value.insert(i+1+j, added)

                        # The above insert call leaves the _raw_data entry as None,
                        # which will cause search indexing to error - Wagtail probably
                        # isn't expecting us to meddle with StreamValues like this.
                        #
                        # Thus we also need to populate the _raw_data entry manually.

                        parent_value._raw_data[i+1+j] = added.get_prep_value()
                    break

            save()

            return ReloadResponse()
        except forms.ValidationError as e:
            errors = ErrorList([e])
            blank_value = val

    bw = BlockWidget(parent_block)

    return render_edit_panel(request, {
        'form_html': bw.render_with_errors('block_edit_form', blank_value, errors=errors),
        'form_media': bw.media,
    })



import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

DEBUG = True

ALLOWED_HOSTS = ['*']

INSTALLED_APPS = [
    'tests',
    'liveedit',

    'wagtail.contrib.forms',
    'wagtail.contrib.redirects',
    'wagtail.contrib.routable_page',
    'wagtail.embeds',
    'wagtail.sites',
    'wagtail.users',
    'wagtail.snippets',
    'wagtail.documents',
    'wagtail.images',
    #'wagtail.search', # produces deprecation warnings
    'wagtail.admin',
    'wagtail.core',

    'modelcluster',
    'taggit',

    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

SECRET_KEY = 'secretkey'

MIDDLEWARE = [
	'django.middleware.security.SecurityMiddleware',
	'django.contrib.sessions.middleware.SessionMiddleware',
	'django.middleware.common.CommonMiddleware',
	'django.middleware.csrf.CsrfViewMiddleware',
	'django.contrib.auth.middleware.AuthenticationMiddleware',
	'django.contrib.messages.middleware.MessageMiddleware',
	'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
	{
		'BACKEND': 'django.template.backends.django.DjangoTemplates',
		'DIRS': ["tests/templates"],
		'APP_DIRS': True,
		'OPTIONS': {
			'context_processors': [
				'django.template.context_processors.debug',
				'django.template.context_processors.request',
				'django.contrib.auth.context_processors.auth',
				'django.contrib.messages.context_processors.messages',
			],
		},
	},
]

ROOT_URLCONF = 'tests.urls'

STATIC_URL = '/static/'

STATIC_ROOT = os.path.join(BASE_DIR, 'static')

WAGTAILADMIN_BASE_URL = "https://example.com/"



# Generated by Django 3.2.10 on 2022-01-03 05:13

from django.db import migrations, models
import django.db.models.deletion
import wagtail.core.blocks
import wagtail.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('body', wagtail.core.fields.StreamField([('text', wagtail.core.blocks.StructBlock([('body', wagtail.core.blocks.RichTextBlock(required=False))], template='text_block.html')), ('section', wagtail.core.blocks.StreamBlock([('text', wagtail.core.blocks.StructBlock([('body', wagtail.core.blocks.RichTextBlock(required=False))], template='text_block.html'))])), ('list', wagtail.core.blocks.StructBlock([('title', wagtail.core.blocks.TextBlock(required=False)), ('items', wagtail.core.blocks.StreamBlock([('item', wagtail.core.blocks.StructBlock([('body', wagtail.core.blocks.RichTextBlock(required=False))], template='text_block.html'))]))]))], blank=True, null=True)),
            ],
            bases=('wagtailcore.page',),
        ),
    ]



def is_enabled(request):
    """
    This is called before the liveedit functionality is inserted into a page.
    You can replace this function by adding the Django setting
    `LIVEEDIT_ENABLED_CHECK` pointing to a function in your own code.

    You may wish to only enable liveedit in certain circumstances by checking a
    condition here. 
    
    You may also want to perform your own admin-only cookie check here, to avoid
    `request.user` being accessed by the subsequent `is_authenticated` check,
    since this will access the session, setting the `Vary: Cookie` header and
    spoiling the cacheability of the page.

    """
    return True



from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^action/', views.action_view),
    re_path(r'^append-block/', views.append_block_view),
    re_path(r'^edit-block/', views.edit_block_view),
]



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.core.servers.basehttp import WSGIServer
from django.db import connections, transaction
from django.test import LiveServerTestCase
from django.test.testcases import LiveServerThread, QuietWSGIRequestHandler

from wagtail.core.models import Page
from wagtail.tests.utils import WagtailPageTests, WagtailTestUtils

from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.proxy import Proxy, ProxyType

import json
import os
import socket
import time
import uuid

from .models import TestPage

hostname = socket.gethostname()
local_ip = socket.gethostbyname(hostname)

class LiveServerSingleThread(LiveServerThread):
    """Runs a single threaded server rather than multi threaded. Reverts https://github.com/django/django/pull/7832"""

    def _create_server(self):
        """
        the keep-alive fixes introduced in Django 2.1.4 (934acf1126995f6e6ccba5947ec8f7561633c27f)
        cause problems when serving the static files in a stream.
        We disable the helper handle method that calls handle_one_request multiple times.
        """
        QuietWSGIRequestHandler.handle = QuietWSGIRequestHandler.handle_one_request

        return WSGIServer((self.host, self.port), QuietWSGIRequestHandler, allow_reuse_address=False)


class LiveServerSingleThreadedTestCase(StaticLiveServerTestCase):
    "A thin sub-class which only sets the single-threaded server as a class"
    server_thread_class = LiveServerSingleThread

class FrontendTest(LiveServerSingleThreadedTestCase, WagtailPageTests, WagtailTestUtils):
    host = '0.0.0.0'
    port = 31411

    def setUp(self):
        super(FrontendTest, self).setUp()

        # user = get_user_model().objects.create(username='testuser', is_staff=True)
        # user.set_password('testpassword')
        # user.save()

        self.live_server_url = 'http://%s:%d'%(local_ip, self.port)

        opts = Options()
        opts.headless = True
        opts.add_argument('--disable-web-security') # avoid CORS problems
        opts.add_argument('--disable-site-isolation-trials')
        self.driver = webdriver.Remote(
            command_executor='http://%s:4444/wd/hub'%os.getenv('SELENIUM_HOST'),
            desired_capabilities=DesiredCapabilities.CHROME,
            options=opts,
            #proxy=proxy
        )
        self.driver.set_window_size(1280, 800)

        self.user = self.create_superuser(
            username='administrator',
            email='administrator@email.com',
            password='password'
        )
        self.root_page = Page.objects.get(pk=1)

        self.test_page = TestPage()
        self.test_page.title = "Test"
        self.test_page.slug = 'test'
        self.test_page.body = json.dumps([
            {'type': 'text', 'id':str(uuid.uuid4()), 'value': {
                'body': "<h2>hello world</h2>"
            }},
            {'type': 'section', 'id':str(uuid.uuid4()), 'value': [
                {'type': 'text', 'id':str(uuid.uuid4()), 'value': {
                    'body': "<p>Text inside a section.</p>"
                }},
                {'type': 'text', 'id':str(uuid.uuid4()), 'value': {
                    'body': "<p>A second text inside a section.</p>"
                }},
            ]},
            {'type': 'list', 'id':str(uuid.uuid4()), 'value': {
                'title': 'This is a list',
                'items': [
                    {'type': 'item', 'id':str(uuid.uuid4()), 'value': {
                        'body': '<p>Testing</p>'
                    }}
                ],
            }},
        ])
        self.root_page.add_child(instance=self.test_page)
        self.content_type = ContentType.objects.get_for_model(TestPage)

        self.driver.implicitly_wait(1) #set timeout for finding elements

        self.driver.get(self.live_server_url + '/admin/?next=/404')
        self.xpath("//input[@name='username']").send_keys('administrator')
        self.xpath("//input[@name='password']").send_keys('password')
        self.xpath("//input[@name='password']").submit()

    def tearDown(self):
        self.driver.quit()
        super(FrontendTest, self).tearDown()

    def sel(self, sel):
        return self.driver.find_element(By.CSS_SELECTOR, sel)

    def xpath(self, sel):
        return self.driver.find_element(By.XPATH, sel)

    def hover(self, el):
        ActionChains(self.driver).move_to_element(el).perform()

    def test_edit(self):
        self.driver.get(self.live_server_url + '/page/')

        self.hover(self.sel('.block-text'))
        self.xpath('//button[text()="Edit"]').click()

        time.sleep(1) #wait for editor to load

        self.driver.switch_to.frame(self.driver.find_elements(By.TAG_NAME, "iframe")[0])

        h2 = self.sel('h2')
        h2.click()
        h2.send_keys('extra_header_text')

        time.sleep(1) #wait for Draftail to update

        self.sel('input[value="Save"]').click()

        time.sleep(1) #wait for save to stick

        h2_contents = TestPage.objects.get(pk=self.test_page.id).body.raw_data[0]['value']['body']
        assert 'extra_header_text' in h2_contents



from django.conf import settings
from django.conf.urls.static import static
from django.urls import path, include

from wagtail.core import urls as wagtail_urls

from .views import page_view

urlpatterns = [
    path('__liveedit__/', include('liveedit.urls')),
    path('admin/', include('wagtail.admin.urls')),
    path('page/', page_view),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='wagtail-liveedit',
      version='0.0.7',
      description='Live editing add-on for Wagtail CMS',
      author='jonny5532',
      license='MIT',
      url='https://github.com/jonny5532/wagtail-liveedit',
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: Wagtail",
      ],
      packages=find_packages(exclude=['tests', 'tests.migrations']),
      include_package_data=True,
      install_requires=[
          'wagtail>=2.13',
      ],
     )
 



from django.shortcuts import render

from .models import TestPage

def page_view(request):
    try:
        page = TestPage.objects.get(slug='test')

        return render(request, "page.html", {'page': page})
    except Exception as e:
        print(e)


from django.contrib.contenttypes.models import ContentType
from django.db import connection, transaction
from django.shortcuts import render
from django.template.loader import render_to_string
from django.test import RequestFactory

from wagtail.core.models import Page
from wagtail.core.rich_text import RichText
from wagtail.tests.utils import WagtailPageTests, WagtailTestUtils
from wagtail.tests.utils.form_data import nested_form_data, streamfield

from html.parser import HTMLParser
import json
import urllib.parse
import uuid

from liveedit import views

from .models import TestPage

class MockRequest:
    def __init__(self, user):
        self.user = user

class BackendTestCase(WagtailPageTests, WagtailTestUtils):
 
    def setUp(self):
        super(BackendTestCase, self).setUp()

        self.user = self.create_superuser(
            username='administrator',
            email='administrator@email.com',
            password='password'
        )
        self.root_page = Page.objects.get(pk=1)

        self.test_page = TestPage()
        self.test_page.title = "Test"
        self.test_page.slug = 'test'
        self.test_page.body = json.dumps([
            {'type': 'text', 'id':str(uuid.uuid4()), 'value': {
                'body': "<h2>hello world</h2>"
            }},
            {'type': 'section', 'id':str(uuid.uuid4()), 'value': [
                {'type': 'text', 'id':str(uuid.uuid4()), 'value': {
                    'body': "<p>Text inside a section.</p>"
                }},
                {'type': 'text', 'id':str(uuid.uuid4()), 'value': {
                    'body': "<p>A second text inside a section.</p>"
                }},
            ]},
            {'type': 'list', 'id':str(uuid.uuid4()), 'value': {
                'title': 'This is a list',
                'items': [
                    {'type': 'item', 'id':str(uuid.uuid4()), 'value': {
                        'body': '<p>Testing</p>'
                    }}
                ],
            }},
        ])
        self.root_page.add_child(instance=self.test_page)
        self.content_type = ContentType.objects.get_for_model(TestPage)

    def test_liveedit_attributes(self):
        ret = render_to_string(
            "page.html", 
            {'page':self.test_page}, 
            request=MockRequest(self.user)
        )

        parent = self
        class MyHTMLParser(HTMLParser):
            def handle_starttag(self, tag, attrs):
                if (tag=="div" 
                        and dict(attrs).get('class')=='block-text'
                        and dict(attrs).get('data-liveedit')):
                    if not hasattr(self, 'data_liveedit'):
                        self.data_liveedit = json.loads(dict(attrs).get('data-liveedit'))

            def check_tags(self):
                parent.assertEqual(self.data_liveedit['id'], parent.test_page.body[0].id)
                parent.assertEqual(self.data_liveedit['block_type'], parent.test_page.body[0].block_type)
                #check content type?
                parent.assertEqual(self.data_liveedit['object_field'], 'body')

        parser = MyHTMLParser()
        parser.feed(ret)
        parser.check_tags()

    def test_block_edit_form(self):
        self.login(user=self.user)

        ret = self.client.get('/__liveedit__/edit-block/', {
            'content_type_id':self.content_type.id,
            'object_id':self.test_page.id,
            'object_field':'body',
            'id':self.test_page.body[1].value[0].id,
        })

        parent = self
        class MyHTMLParser(HTMLParser):
            def handle_starttag(self, tag, attrs):
                if (tag=="div" 
                        and dict(attrs).get('id')=='block_edit_form'
                        and dict(attrs).get('data-block')):
                    self.data_block = json.loads(dict(attrs).get('data-block'))

            def check_tags(self):
                assert isinstance(self.data_block, dict)
                assert len(self.data_block)

        parser = MyHTMLParser()
        parser.feed(ret.content.decode('utf-8'))
        parser.check_tags()

    def test_block_edit(self):
        self.login(user=self.user)

        ret = self.client.post('/__liveedit__/edit-block/?' + urllib.parse.urlencode({
            'content_type_id':self.content_type.id,
            'object_id':self.test_page.id,
            'object_field':'body',
            'id':self.test_page.body[0].id,
        }), {
            'block_edit_form-body':json.dumps({
                "blocks":[{
                    "text":"This is the replacement rich text.",
                }],
            })
        })

        #print(str(TestPage.objects.get(pk=self.test_page.id).body))
        assert '<p>This is the replacement rich text.</p>' in str(TestPage.objects.get(pk=self.test_page.id).body)

    def _do_action(self, block_id, action):
        self.login(user=self.user)

        ret = self.client.post('/__liveedit__/action/', {
            'content_type_id':self.content_type.id,
            'object_id':self.test_page.id,
            'object_field':'body',
            'id':block_id,
            'redirect_url':'/test/#le-' + block_id,
            'action':action,
        })

        return TestPage.objects.get(pk=self.test_page.id).body

    def test_block_move_down(self):
        block_id = self.test_page.body[0].id
        body = self._do_action(block_id, 'move_down')

        self.assertEqual(
            body[1].id,
            block_id,
        )

    def test_block_move_up(self):
        block_id = self.test_page.body[1].id
        body = self._do_action(block_id, 'move_up')

        self.assertEqual(
            body[0].id,
            block_id,
        )

    def test_block_delete(self):
        self.login(user=self.user)

        n = len(self.test_page.body)

        ret = self.client.post('/__liveedit__/edit-block/?' + urllib.parse.urlencode({
            'content_type_id':self.content_type.id,
            'object_id':self.test_page.id,
            'object_field':'body',
            'id':self.test_page.body[1].id,
        }), {
            'delete':'1'        
        })

        self.assertEqual(
            len(TestPage.objects.get(pk=self.test_page.id).body),
            n-1
        )

    def test_block_delete_nested(self):
        self.login(user=self.user)

        n = len(self.test_page.body[1].value)

        ret = self.client.post('/__liveedit__/edit-block/?' + urllib.parse.urlencode({
            'content_type_id':self.content_type.id,
            'object_id':self.test_page.id,
            'object_field':'body',
            'id':self.test_page.body[1].value[0].id,
        }), {
            'delete':'1'        
        })

        self.assertEqual(
            len(TestPage.objects.get(pk=self.test_page.id).body[1].value),
            n-1
        )



from wagtail.core import blocks
from wagtail.core.fields import StreamField
from wagtail.core.models import Page

class TestPage(Page):
    #class Meta:
        #app_label = 'tests'
        #managed = False

    template = 'page.html'

    body = StreamField([

        ('text', blocks.StructBlock([
            ('body', blocks.RichTextBlock(required=False)),
        ], template="text_block.html")),

        ('section', blocks.StreamBlock([
            ('text', blocks.StructBlock([
                ('body', blocks.RichTextBlock(required=False)),
            ], template="text_block.html")),
        ])),

        ('list', blocks.StructBlock([
            ('title', blocks.TextBlock(required=False)),
            ('items', blocks.StreamBlock([
                ('item', blocks.StructBlock([
                    ('body', blocks.RichTextBlock(required=False)),
                ], template="text_block.html")),
            ])),
        ])),

    ], null=True, blank=True)

